#!/usr/bin/env python
#
# Copyright (C) 2014 - Renzo Davoli
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# 

import sys, copy, json, argparse, os

inch=2.54
#inch=1

class mille:
	'''
	pcb object: size in mm, width/height in .1" elements.
	table: the same integer represents the same track
	drill: boolean (in the future values can represent different diameters)
	rear: for two sided pcb, not yet implemented
	'''
	def __init__(self):
		self.size=(0,0)
		self.width,self.height=0,0
		self.table=[]
		self.rear=[]
		self.drill=[]
	def valid(self,i,j):
		return True if i>0 and j>0 and i<self.width+1 and j<self.height+1 else False
	def new(self,size=(120,100),twosided=0):
		self.width,self.height=size
		self.size=size
		self.width,self.height=int(self.width//inch),int(self.height//inch)
		self.drill=[[0 for _ in range(self.width+2)] for _ in range(self.height+2)]
		self.table=[[1 if self.valid(i,j) else 0 for i in range(self.width+2)]
			for j in range(self.height+2)]
		if twosided:
			self.rear=[[1 if self.valid(i,j) else 0 for i in range(self.width+2)]
				for j in range(self.height+2)]
	def reversei(self):
		for i in range((self.height+2)//2):
			self.table[i],self.table[self.height+1-i]=self.table[self.height+1-i],self.table[i]
			self.drill[i],self.drill[self.height+1-i]=self.drill[self.height+1-i],self.drill[i]
			if self.rear:
				self.rear[i],self.rear[self.height+1-i]=self.rear[self.height+1-i],self.rear[i]
	def reversej(self):
		for i in range(self.height+2):
			for j in range((self.width+2)//2):
				self.table[i][j],self.table[i][self.width+1-j]=self.table[i][self.width+1-j],self.table[i][j]
				self.drill[i][j],self.drill[i][self.width+1-j]=self.drill[i][self.width+1-j],self.drill[i][j]
				if self.rear:
					self.rear[i][j],self.rear[i][self.width+1-j]=self.rear[i][self.width+1-j],self.rear[i][j]

xoff,yoff=0,0

class defargs:
	uplevel=2.0
	downlevel=-0.1
	zspeed=50
	g0speed=2000
	g1speed=100

def loadrc():
	try:
		with open(os.path.expanduser("~/.milleforirc"), "r") as f:
			loaded=json.load(f)
			for f in loaded:
				defargs.__dict__[f]=loaded[f]
	except IOError:
		pass

def main():
	pcb=mille()
	loadrc()
	parser = argparse.ArgumentParser()
	parser.add_argument("input_file", help="millefori input file")
	parser.add_argument("-u", "--uplevel", type=float, help="traveling z-level ({})".format(defargs.uplevel), default=defargs.uplevel)
	parser.add_argument("-d", "--downlevel", type=float, help="milling z-level ({})".format(defargs.downlevel), default=defargs.downlevel)
	parser.add_argument("-z", "--zspeed", type=int, help="drilling z speed ({})".format(defargs.zspeed), default=defargs.zspeed)
	parser.add_argument("--g0speed", type=int, help="moving speed ({})".format(defargs.g0speed), default=defargs.g0speed)
	parser.add_argument("--g1speed", type=int, help="working speed ({})".format(defargs.g1speed), default=defargs.g1speed)
	parser.add_argument("-v", "--verbose", action="store_true", help="verbose output (add comments)")
	parser.add_argument("-R", "--round", action="store_true", help="round corners of PCB tracks using G2/G3")
	parser.add_argument("-r", "--emuround", action="store_true", help="emulated round corners of PCB tracks")
	parser.add_argument("-1", "--no1", action="store_true", help="skip track 1")
	parser.add_argument("-2", "--backside", action="store_true", 
			help="backside (component side)")
	parser.add_argument('--xyflip', action="store_true", 
			help='backside: flip x and y, implies -2')
	parser.add_argument('--centered', action="store_true", help='xy centered')
	args = parser.parse_args()
	args.speed=[args.g0speed,args.g1speed]
	pcb.filename=args.input_file
	if file_load(pcb) == False:
		print "Error opening file", filename
	if args.xyflip: args.backside=True
	sx,sy=pcb.size
	if args.centered:
		args.xoff=max((sx-pcb.width*inch)/2,0)
		args.yoff=max((sy-pcb.height*inch)/2,0)
	else:
		args.xoff=args.yoff=0
	pcb.reversei()
	if args.backside:
		if not args.xyflip:
			args.xoff=sx-args.xoff
		table=pcb.rear
	else:
		table=pcb.table
	pcb.reversej()
	tablerenumber(pcb,table)
	if args.verbose:
		if args.backside:
			msg="; PCB BACK SIDE (COMPONENT): "
			if args.xyflip:
				msg+="XY FLIPPED: ROTATE THE PCB AROUND THE ORIGIN (0,0)"
			else:
				msg+="ROTATE THE PCB HORIZONTALLY"
		else:
			msg="; PCB SOLDERING SIDE"
		print msg
	computepath(pcb,table,args,0,0)

# GCODE COMMANDS
def up(args):
	print "G1 Z{} F{}".format(args.uplevel,args.zspeed)

def down(args):
	print "G1 Z{} F{}".format(args.downlevel,args.zspeed)

def gg(args,g,x,y):
	if args.xyflip:
		xx,yy=y*inch+args.yoff,x*inch+args.xoff
	elif args.backside:
		xx,yy=args.xoff-x*inch,y*inch+args.yoff
	else:
		xx,yy=x*inch+args.xoff,y*inch+args.yoff
	print "G{} X{} Y{} F{}".format(g,xx,yy,args.speed[g])

def g23(args,g,x,y,dx,dy):
	if args.xyflip:
		xx,yy=y*inch+args.yoff,x*inch+args.xoff
		dx,dy=dy*inch,dx*inch
		g=3 if g==2 else 2
	elif args.backside:
		xx,yy=args.xoff-x*inch,y*inch+args.yoff
		dx,dy=-dx*inch,dy*inch
		g=3 if g==2 else 2
	else:
		xx,yy=x*inch+args.xoff,y*inch+args.yoff
		dx,dy=dx*inch,dy*inch
	print "G{} X{} Y{} I{} J{} F{}".format(g,xx,yy,dx,dy,args.speed[1])

'''
renumber the table: 
each connected component gets a different id
'''
def tablerenumber(pcb,table):
	pcb.remap={}
	counter={}
	def minfree():
		color=1;
		while True:
			if counter.get(color,0)==0:
				break
			color=color+1
		counter[color]=0
		return color
	def redefine(x,y):
		color=table[y][x]
		if not pcb.remap.get(color,False):
			newcolor=minfree()
			pcb.remap[newcolor]=color
			l=[(x,y)]
			while len(l) > 0:
				(x,y)=l.pop()
				if table[y][x]==color:
					table[y][x]=newcolor
					counter[color]-=1
					counter[newcolor]+=1
					l.append((x,y+1))
					l.append((x,y-1))
					l.append((x+1,y))
					l.append((x-1,y))
	for i in range(1,pcb.width+1):
		for j in range(1,pcb.height+1):
			color=table[j][i]
			counter[color]=counter.get(color,0)+1
	for i in range(1,pcb.width+1):
		for j in range(1,pcb.height+1):
			redefine(i,j)

'''
get the nearest non "printed" track.
and then get its lower left corner
'''
def nearest(pcb,table,x,y):
	def square(x): return x*x
	mini,minj=0,0
	mind=(square(pcb.width)+square(pcb.height)) ** 0.5
	for i in range(1,pcb.width+1):
		for j in range(1,pcb.height+1):
			if table[j][i] != 0:
				dx,dy=x-i,y-j
				d=abs(dx)+abs(dy)
				if d<mind:
					mini,minj=i,j
					mind=d
	if not (mini,minj) == (0,0):
		color=table[minj][mini]
		mini,minj=0,0
		mind=abs(pcb.width)+abs(pcb.height)
		for i in range(1,pcb.width+1):
			for j in range(1,pcb.height+1):
				if table[j][i] == color:
					d=((i*i)+(j*j)) ** 0.5
					if d<mind:
						mini,minj=i,j
						mind=d
	return mini,minj

'''
tag the connected component as "done"
'''
def zerostep(pcb,table,x,y):
	color=table[y][x]
	l=[(x,y)]
	while len(l) > 0:
		(x,y)=l.pop()
		if table[y][x] == color:
			table[y][x]=0
			l.append((x,y+1))
			l.append((x,y-1))
			l.append((x+1,y))
			l.append((x-1,y))

dirs=[(1,0),(0,1),(-1,0),(0,-1)]
ends=([(0,-1),(0,0),(-1,0),(-1,-1)], [(-0.5,-1),(0,-0.5),(-0.5,0),(-1,-0.5)])

nextdir=[1,2,3,0]
prevdir=[3,0,1,2]

tsc=[(0.25881904510252074, 0.03407417371093169), 
	(0.5, 0.1339745962155613), (0.7071067811865475, 0.2928932188134524), 
	(0.8660254037844386, 0.5), (0.9659258262890683, 0.7411809548974793), (1.0, 1.0)]

def emug23(d, xin, yin, dx, dy, args):
	if d % 2:
		for s,c in tsc:
			gg(args,1,xin+dx*s,yin+dy*c);
	else:
		for s,c in tsc:
			gg(args,1,xin+dx*c,yin+dy*s);

#import time
'''
write the gcode to mill the border of a track
it follows the border counterclockwise starting 
from the lowest-left corner of the track
'''
def pathstep(pcb,table,args,x,y):
	round=1 if args.round or args.emuround else 0
	gg(args,0,x-(0.5 if round else 1),y-1)
	down(args)
	color=table[y][x]
	xx,yy=x,y
	direction=0
	while True:
#		print "pathstep",xx,yy,direction
#		time.sleep(0.1)
		if direction==3 and xx==x and yy==y:
			ex,ey=ends[round][direction]
			gg(args,1,xx+ex,yy+ey);
			direction=nextdir[direction]
			if round:
				rex,rey=ends[round][direction]
				dex,dey=dirs[direction]
				if args.emuround:
					emug23(direction,xx+ex,yy+ey,rex-ex,rey-ey,args)
				else:
					g23(args,3,xx+rex,yy+rey,dex*0.5,dey*0.5);
			break
		dx,dy=dirs[direction]
#		print "NEXTinline",xx+dx,yy+dy,table[yy+dy][xx+dx]
		if table[yy+dy][xx+dx] != color:
			ex,ey=ends[round][direction]
			gg(args,1,xx+ex,yy+ey);
			direction=nextdir[direction]
			if round:
				rex,rey=ends[round][direction]
				dex,dey=dirs[direction]
				if args.emuround:
					emug23(direction,xx+ex,yy+ey,rex-ex,rey-ey,args)
				else:
					g23(args,3,xx+rex,yy+rey,dex*0.5,dey*0.5);
		else:
			rx,ry=dirs[prevdir[direction]]
#			print "NEXTright",xx+rx,yy+ry,table[yy+ry][xx+rx]
			if table[yy+dy+ry][xx+dx+rx] == color:
				ex,ey=ends[round][direction]
				gg(args,1,xx+ex,yy+ey);
				direction=prevdir[direction]
				oldx,oldy=xx,yy
				xx,yy=xx+dx+rx,yy+dy+ry
				if round:
					rex,rey=ends[round][direction]
					dex,dey=dirs[direction]
					if args.emuround:
						oldx,oldy=oldx+ex,oldy+ey
						newx,newy=xx+rex, yy+rey
						emug23(direction,oldx,oldy,newx-oldx,newy-oldy,args)
					else:
						g23(args,2,xx+rex,yy+rey,dex*0.5,dey*0.5);
			else:
				xx,yy=xx+dx,yy+dy
	up(args)

'''
produce the gcode to mill the pcb:
while there are tracks to mill:
	find the nearest unmilled track,
	mill it
	tag it as milled
'''
def computepath(pcb,table,args,x,y):
	up(args)
	while True:
		x,y=nearest(pcb,table,x,y)
		if x == 0:
			break
		if args.verbose:
			print "; track color {} @ {},{}".format(pcb.remap[table[y][x]],
						pcb.width+1-x,pcb.height+1-y) 
		if args.no1:
			if pcb.remap[table[y][x]] > 1:
				pathstep(pcb,table,args,x,y)
		else:
			pathstep(pcb,table,args,x,y)
		zerostep(pcb,table,x,y)

def usage():
	print "Usage:"
	print "  "+sys.argv[0]+" filename"
	sys.exit(1)

def file_load(pcb):
	try:
		try:
			name=pcb.filename
			with open(name, "r") as f:
				loaded=json.load(f)
				for f in loaded:
					pcb.__dict__[f]=loaded[f]
				pcb.filename=name
		except TypeError:
#old format
			with open(pcb.filename, "r") as f:
				pcb.size,pcb.table,pcb.drill=json.load(f)
				pcb.height=len(pcb.table)-2
				pcb.width=len(pcb.table[0])-2
	except IOError:
		return False
	return True

if __name__ == '__main__':
  main()

